@@ -12,8 +12,10 @@ import orjson
 from datetime import datetime
 import copy
 from multiprocessing.pool import ThreadPool
-# import shapely
-from tethysts.utils import get_object_s3, result_filters, process_results_output, read_json_zstd, key_patterns
+from tethysts.utils import get_object_s3, result_filters, process_results_output, read_json_zstd, key_patterns, get_nearest_station, get_intersected_stations, spatial_query
+from shapely.geometry import Point, Polygon, shape
+from typing import Optional, List, Any, Union
+from enum import Enum
 
 pd.options.display.max_columns = 10
 
@@ -26,6 +28,9 @@ pd.options.display.max_columns = 10
 ##############################################
 ### Class
 
+# class Output(str, Enum):
+#     zstd = 'zstd'
+
 
 class Tethys(object):
     """
@@ -65,7 +70,7 @@ class Tethys(object):
             pass
 
 
-    def get_datasets(self, remotes_list, threads=30):
+    def get_datasets(self, remotes_list: List[dict], threads: int = 30):
         """
         The function to get datasets from many remotes.
 
@@ -91,7 +96,7 @@ class Tethys(object):
         return self.datasets
 
 
-    def get_remote_datasets(self, remote):
+    def get_remote_datasets(self, remote: dict):
         """
         Get datasets from an individual remote. Saves result into the object.
 
@@ -127,7 +132,13 @@ class Tethys(object):
             print('No datasets.json.zst file in S3 bucket')
 
 
-    def get_stations(self, dataset_id, results_object_keys=False):
+    def get_stations(self,
+                     dataset_id: str,
+                     geometry: Optional[dict] = None,
+                     lat: Optional[float] = None,
+                     lon: Optional[float] = None,
+                     distance: Optional[float] = None,
+                     results_object_keys: Optional[bool] = False):
         """
         Method to return the stations associated with a dataset.
 
@@ -135,6 +146,14 @@ class Tethys(object):
         ----------
         dataset_id : str
             The dataset_id of the dataset.
+        geometry : dict or None
+            A geometry in GeoJSON format. Can be either a point or a polygon. If it's a point, then the method will perform a nearest neighbor query and return one station. If it's a polygon, then the method performs an intersection of all stations within the polygon.
+        lat : float or None
+            Instead of using the geometry parameter, optionally use lat and lon for the spatial queries. Both lat and lon must be passed for the spatial queries and will override the geometry parameter. If only lat and lon are passed, then the method performs a nearest neighbor query. If distance is passed in addition to lat and lon, then distance is used as a radius buffer and an intersection is performed.
+        lon : float or None
+            See lat description.
+        distance : float or None
+            See lat description. This should be in decimal degrees not meters.
         results_object_keys : bool
             Shoud the results object keys be returned? The results object keys list the available results in Tethys.
 
@@ -147,25 +166,35 @@ class Tethys(object):
 
         site_key = self._key_patterns['stations'].format(dataset_id=dataset_id)
 
-        try:
-            stn_obj = get_object_s3(site_key, remote['connection_config'], remote['bucket'])
-            stn_list = read_json_zstd(stn_obj)
-            stn_list = [s for s in stn_list if isinstance(s, dict)]
+        if dataset_id in self._stations:
+            stn_dict = copy.deepcopy(self._stations[dataset_id])
+        else:
+            try:
+                stn_obj = get_object_s3(site_key, remote['connection_config'], remote['bucket'])
+                stn_list = read_json_zstd(stn_obj)
+                stn_dict = {s['station_id']: s for s in stn_list if isinstance(s, dict)}
 
-            self._stations.update({dataset_id: {s['station_id']: s for s in copy.deepcopy(stn_list)}})
+                self._stations.update({dataset_id: copy.deepcopy(stn_dict)})
 
-            # TODO: Run spatial query here!
+            except:
+                print('No stations.json.zst file in S3 bucket')
+                return None
 
-            if not results_object_keys:
-                s = [s.pop('results_object_key') for s in stn_list]
+        ## Spatial query
+        stn_ids = spatial_query(stn_dict, geometry, lat, lon, distance)
 
-            return stn_list
+        if isinstance(stn_ids, list):
+            stn_list1 = [stn_dict[s] for s in stn_ids]
+        else:
+            stn_list1 = list(stn_dict.values())
 
-        except:
-            print('No stations.json.zst file in S3 bucket')
+        if not results_object_keys:
+            s = [s.pop('results_object_key') for s in stn_list1]
+
+        return stn_list1
 
 
-    def get_run_dates(self, dataset_id, station_id):
+    def get_run_dates(self, dataset_id: str, station_id: str):
         """
         Function to get the run dates of a particular dataset and station.
 
@@ -190,7 +219,7 @@ class Tethys(object):
         return run_dates
 
 
-    def _get_results_obj_key_s3(self, dataset_id, station_id, run_date):
+    def _get_results_obj_key_s3(self, dataset_id: str, station_id: str, run_date: Union[str, pd.Timestamp]):
         """
 
         """
@@ -222,7 +251,18 @@ class Tethys(object):
         return obj_key
 
 
-    def get_results(self, dataset_id, station_id, from_date=None, to_date=None, from_mod_date=None, to_mod_date=None, modified_date=False, quality_code=False, run_date=None, remove_height=False, output='DataArray'):
+    def get_results(self,
+                    dataset_id: str,
+                    station_id: str,
+                    from_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    to_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    from_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    to_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    modified_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    quality_code: Optional[bool] = False,
+                    run_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    remove_height: Optional[bool] = False,
+                    output: str = 'Dataset'):
         """
         Function to query the time series data given a specific dataset_id and station_id. Multiple optional outputs.
 
@@ -284,7 +324,116 @@ class Tethys(object):
         return output1
 
 
-    def get_bulk_results(self, dataset_id, station_ids, from_date=None, to_date=None, from_mod_date=None, to_mod_date=None, modified_date=False, quality_code=False, run_date=None, remove_height=False, output='DataArray', threads=30):
+    def get_nearest_results(self,
+                    dataset_id: str,
+                    geometry: Optional[dict] = None,
+                    lat: Optional[float] = None,
+                    lon: Optional[float] = None,
+                    from_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    to_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    from_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    to_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    modified_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    quality_code: Optional[bool] = False,
+                    run_date: Union[str, pd.Timestamp, datetime, None] = None,
+                    remove_height: Optional[bool] = False,
+                    output: str = 'Dataset'):
+        """
+        Function to query the time series data given a specific dataset_id and 2D point. A nearest neighbor query will be performed to find the single nearest station to the input point. Multiple optional outputs.
+
+        Parameters
+        ----------
+        dataset_id : str
+            The dataset_id of the dataset.
+        geometry : dict or None
+            A geometry in GeoJSON format. Can be either a point or a polygon. If it's a point, then the method will perform a nearest neighbor query and return one station.
+        lat : float or None
+            Instead of using the geometry parameter, optionally use lat and lon for the spatial queries. Both lat and lon must be passed for the spatial queries and will override the geometry parameter. If only lat and lon are passed, then the method performs a nearest neighbor query.
+        lon : float or None
+            See lat description.
+        from_date : str, Timestamp, datetime, or None
+            The start date of the selection.
+        to_date : str, Timestamp, datetime, or None
+            The end date of the selection.
+        from_mod_date : str, Timestamp, datetime, or None
+            Only return data post the defined modified date.
+        to_mod_date : str, Timestamp, datetime, or None
+            Only return data prior to the defined modified date.
+        modified_date : bool
+            Should the modified dates be returned if they exist?
+        quality_code : bool
+            Should the quality codes be returned if they exist?
+        run_date : str or Timestamp
+            The run_date of the results to be returned. Defaults to None which will return the last run date.
+        remove_height : bool
+            Should the height dimension be removed from the output?
+        output : str
+            Output format of the results. Options are:
+                Dataset - return the entire contents of the netcdf file as an xarray Dataset,
+                DataArray - return the requested dataset parameter as an xarray DataArray,
+                Dict - return a dictionary of results from the DataArray,
+                json - return a json str of the Dict.
+
+        Returns
+        -------
+        Whatever the output was set to.
+        """
+        ## Get parameters
+        dataset = self._datasets[dataset_id]
+        parameter = dataset['parameter']
+        remote = self._remotes[dataset_id]
+
+        ## Get all stations
+        if dataset_id not in self._stations:
+            stns = self.get_stations(dataset_id)
+
+        stn_dict = self._stations[dataset_id]
+
+        ## Spatial query
+        # Checks
+        if isinstance(geometry, dict):
+            geom_type = geometry['type']
+        else:
+            geom_type = None
+
+        if not ((geom_type == 'Point') or (isinstance(lat, float) and isinstance(lon, float))):
+            raise ValueError('geometry or a combination of lat and lon must be passed.')
+
+        # Run the spatial query
+        station_id = spatial_query(stn_dict, geometry, lat, lon)[0]
+
+        ## Get object key
+        obj_key = self._get_results_obj_key_s3(dataset_id, station_id, run_date)
+
+        ## Get results
+        ts_obj = get_object_s3(obj_key.iloc[0], remote['connection_config'], remote['bucket'], 'zstd')
+        xr3 = xr.open_dataset(ts_obj)
+
+        ## Filters
+        ts_xr1 = result_filters(xr3, from_date, to_date, from_mod_date, to_mod_date, remove_height)
+
+        if not 'station_id' in list(ts_xr1.coords):
+            ts_xr1 = ts_xr1.expand_dims('station_id').set_coords('station_id')
+
+        ## Output
+        output1 = process_results_output(ts_xr1, parameter, modified_date, quality_code, output)
+
+        return output1
+
+
+    def get_bulk_results(self,
+                         dataset_id: str,
+                         station_ids: List[str],
+                         from_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         to_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         from_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         to_mod_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         modified_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         quality_code: Optional[bool] = False,
+                         run_date: Union[str, pd.Timestamp, datetime, None] = None,
+                         remove_height: Optional[bool] = False,
+                         output: str = 'Dataset',
+                         threads: int = 30):
         """
         Function to bulk query the time series data given a specific dataset_id and a list of station_ids. The output will be specified by the output parameter and will be concatenated along the station_id dimension.
 
